"""
Richard Gomez juillet 2017

version 3 : pour Benjamin Clerc

Note. permutations du module itertools est utilisée dans isomorphe.

Note. clock du module time est utilisée dans chrono.

Note. On devrait pouvoir améliorer la fonction nommée pseudo_associative...

Note. grace à sudoku_bis, on trouve les groupes d'ordre 7 à isomorphisme
      près en 2.612 secondes (il n'y en a qu'un). 
"""


from itertools import permutations
from time import clock
from random import shuffle

def chrono(fonction):
    """Décorateur de fonction. Chaque fonction décorée avec chrono
    affichera son temps d'exécution à chaque appel.
    """
    def wrapper(*args,**kwargs):
        t = clock()
        y = fonction(*args,**kwargs)
        print("Fonction",fonction.__name__,":",clock()-t,"secondes")
        return y
    return wrapper



@chrono
def sudoku(n):
    """ n : int
    sudoku(n) --> retourne la liste des sudokus n x n, voir définition 6
    """
    GC = [[list(range(n))]]    # GC est une liste de matrices (une matrice est une liste de listes)
    for i in range(1,n):
        GN = []
        for M in GC:
            DM = []    # DM contiendra les presque-sudokus enfantés par M
            gc = [[i]]
            for j in range(1,n):
                gn = []
                for L in gc:
                    DL = []    # DL contiendra les presque-lignes enfantées par L
                    support = set(range(n)) - set(L) - {M[i][j] for i in range(0,i)}
                    for k in support:
                        DL.append(L+[k])    # on concatène avec append, c'est plus rapide et économe en mémoire                       
                    gn = gn + DL
                gc = gn
            for L in gc:
                DM.append(M+[L])
            GN = GN + DM
        GC = GN
        # print("Génération",i,":",len(GN),"presque-sudokus.")  
    return GC

def accepte(M,L):
    n = len(M[0])
    m = len(M)
    for j in range(n):
        if L[j] in [M[i][j] for i in range(m)]:
            return False
    return True 





def zeros(M):
    """La fonction zeros est utilisée par la fonction sudoku_bis (Cf. plus bas).
    M est un presque-sudoku. Par conséquent, dans chaque ligne de M, il y a exactement un zéro.
    zeros(M) retourne la liste des positions des zéros de M.
    Exemple : si M = [[0,1,2,3,4],
                      [1,4,0,2,3],
                      [2,3,4,0,1]]
    alors zeros(M) retourne [0, 2, 3].
    """
    z = []
    for L in M:
        position = L.index(0)    # position de 0 dans L
        z.append(position)
    return z


def pseudo_associative(M):
    """La fonction pseudo_associative est utilisée par la fonction sudoku_bis (Cf. plus bas).
    M est un presque-sudoku m x n, ainsi, M définit "presque" une loi interne dans
    {0,1,...,n-1}.
    Cette fonction retourne True si pour tous i, j, k tels que (ij)k et i(jk) soient définis
    on a bien l'égalité (ij)k = i(jk). False sinon.
    Note. On fera (si c'est possible) les calculs (ij)k et i(jk).
    """
    m = len(M)
    n = len(M[0])   
    for i in range(1,m):       # limite imposée par le calcul i x j
        for j in range(1,m):   # limite imposée par le calcul j x k  
            if M[i][j]<m:      # limite imposée par le calcul (ij) k
                for k in range(1,n):
                    if M[j][k]<m and M[M[i][j]][k]!=M[i][M[j][k]]:
                        return False
    return True


@chrono
def sudoku_bis(n):
    """sudoku_bis est une amélioration de sudoku.
    Nous avons apporté deux améliorations :
    - une amélioration basée sur une observation sur les zéros ;
    - une amélioration apportée par la fonction nommée pseudo_associative.
    Note. sudoku_bis(n) ne retourne que des groupes (avec multiplicité).
    On est obligé en aval
    de trier les groupes retournés pour ne garder qu'un groupe par classe
    d'isomorphisme.
    L'appel sudoku_bis(7) trouve 120 groupes en 2.604 secondes.
    Les 120 groupes sont isomorphes entre eux.
    Python met 2.604 + 0.008 secondes pour trouver l'unique groupe d'ordre 7
    à isomoprhisme près.
    """
    GC = [[list(range(n))]]       
    for i in range(1,n):
        GN = []
        for M in GC:
            z = zeros(M)
            DM = []  
            gc = [[i]]
            for j in range(1,n):
                gn = []
                for L in gc:
                    if i in z and z.index(i)==j:
                        DL = [L+[0]]   # la symétrie des zéros impose que L n'enfante qu'une presque-ligne
                    else:
                        DL = []
                        support = set(range(n)) - set(L) - {M[i][j] for i in range(0,i)}
                        if 0 in support and i>j and M[j][i]!=0:
                            support.remove(0)   # 0 n'est pas accepté parce qu'il casse la symétrie des zéros
                        for k in support:
                            DL.append(L+[k])
                    gn = gn + DL
                gc = gn
            for L in gc:
                fils = M + [L]
                if pseudo_associative(fils):
                    DM.append(fils)                                                                   
            GN = GN + DM                   
        GC = GN
        # print("Génération",i,":",len(GN),"presque-sudokus.")  
    return GC





def associative(M):
    """M est sudoku n x n, càd une table de Pythagore.
    La fonction retourne True si la loi définie par M est associative.
    """
    n = len(M) 
    for a in range(n):
        for b in range(n):
            for c in range(n):
                if M[M[a][b]][c]!=M[a][M[b][c]]:
                    return False
    return True





def isomorphisme(f,M,K):
    """M et K sont des matrices n x n.
    Chacune de ces matrices représente une loi de groupe dans {0,...,n-1}.
    f est une application de {0,...,n-1} vers {0,...,n-1} (bijective).
    f est encodée par un uplet.
    isomorphisme(f,M,K) retourne True si f est un isom de groupes.
    """
    n = len(M)   
    for a in range(1,n):
        for b in range(1,n):
            if f[M[a][b]]!=K[f[a]][f[b]]:
                return False
    return True


def isomorphe(M,K):
    """M et K sont des matrices n x n.
    Chacune de ces matrices représente une loi de groupe dans {0,...,n-1}.
    isomorphe(M,K) renvoie True si les deux groupes sont isomorphes.
    Plus précisément : elle retourne un isomorphisme (s'il en existe un).
    """
    n = len(M)
    for f in permutations(range(1,n)):  # inutile de faire bouger 0
        f = (0,) + f                        # on encode que f(0)=0
        if isomorphisme(f,M,K):
            return f
    return False


def est_dans(M,LL):
    """D'un point de vue mathématique, M est un groupe et
    LL est une liste de groupes.
    La fonction renvoie True si M est isomorphe à un item de LL.
    """
    for K in LL:
        if isomorphe(M,K):
            return True
    return False


@chrono
def groupes(n):
    """groupes(n) retourne la liste exhaustive des groupes d'ordre n à isom près.
    groupes utilise sudoku.
    """
    S = sudoku(n)   
    G = []     # G sera la liste des résultats
    for M in S:
        if associative(M) and not est_dans(M,G):
            G.append(groupe)
    return G



@chrono
def groupes_bis(n):
    """groupes_bis(n) retourne la liste exhaustive des groupes d'ordre n à isom près.
    groupe_bis est une amélioration de groupes : au lieu d'utiliser sudoku, elle utilise sudoku_bis.
    """
    S = sudoku_bis(n)    # tous les items de S sont des groupes : pas besoin de tester l'associativité.
    G = []  
    for M in S:
        if not est_dans(M,G):                       
            G.append(groupe)
    return G

def permutation_alea(n):
	S = list(range(n))  # S reçoit [0,...,n-1]
	shuffle(S)   # mélange les items de S
	return S

def sudoku_alea(n):
	M = [list(range(n))]
	for i in range(1,n):
		continuer = True
		while continuer:
			L = permutation_alea(n)
			L.remove(i)
			ligne = [i] + L
			if accepte(M,ligne):
				continuer = False
		M = M + [ligne]
	return M

def groupe_alea(n):
	while 1:
		M = sudoku_alea(n)
		if associative(M):
			return M
		    
S=[[]]*7
for n in range(2,7):
	for i in range(30):
		M=groupe_alea(n)
		if M not in S[n] and not est_dans(M,S[n]):
			S[n]=S[n]+[M]
print(S)


    
